import Base from './base';
import { onClickOutsideUp } from '../ui_utils'
import { getClickPoint, createSvg, createElement, getActualPoint, getInitialPoint } from './utils';
import { MARGIN_DISTANCE } from '../../constants'

export default class Circle extends Base {
  constructor ({
    annotation,
    layer,
    show = false,
    enableReply = true
  }) {
    super({
      annotation,
      layer,
    })
    this.show = show
    this.enableReply = enableReply

    this.hidden = true
    this.initial = false

    this.start = null
    this.end = null
    this.rect = null
    this.realLinePoints = null

    this.svgElement = null

    this.ratio = window.devicePixelRatio || 1
    this.decimalPlaces = 0 // 保留小数位
    this.isPopupShow = false

    this.onClick = this.handleClick.bind(this)
    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    this.onOpenReply = this.openReply.bind(this)
    this.handlePropertyPanel = this.handlePropertyPanel.bind(this)
    this.onHandlePopup = this.handlePopup.bind(this)
    this.onKeydown = this.handleKeyDown.bind(this)
    
    this.render()
  }

  async render () {
    const annotation = this.annotation
    const scale = this.layer.viewport.scale
    this.updateRect()
    const rect = this.rect
    this.realLinePoints = this.annotation.linePoints.map(num => num * scale)

    if (annotation.precision) {
      if (annotation.precision > 1) {
        annotation.precision = Math.pow(10, -Math.floor(Math.log10(annotation.precision)))
      }
      this.decimalPlaces = Math.abs(annotation.precision).toString().split('.')[1]?.length || 0
    }
    if (!annotation.borderColor) {
      annotation.borderColor = annotation.color || '#FF0000'
    }
    if (!annotation.factor) {
      annotation.factor = await this.layer.documentViewer.messageHandler.sendWithPromise('GetFactor', annotation.rulerPtr)
    }
    if (!annotation.fontStyle) {
      annotation.fontStyle = 'regular'
    }
    if (!annotation.contents) {
      this.getResult()
    }

    this.annotationContainer = createElement(
      'div',
      {
        top: rect.top + 'px',
        left: rect.left + 'px',
        width: rect.width + 'px',
        height: rect.height + 'px'
      }, {
        id: annotation.name,
        class: 'annotation'
      }
    )
    this.svgElement = createSvg(
      'svg',
      {
        viewBox: `0 0 ${rect.width} ${rect.height}`
      }, {
        pointerEvents: 'all',
        overflow: 'visible',
        opacity: annotation.opacity ? annotation.opacity : ''
      }
    )
    this.annotationContainer.append(this.svgElement)
    this.svgElement.addEventListener('mousedown', this.onClick)
    this.layer.div.append(this.annotationContainer)
    this.draw()

    this.outerLineContainer = createElement('div', {}, { class: 'outline-container' })
    this.outerLine = createSvg('svg', {}, {
      position: 'absolute',
      left: `${rect.left - 7}px`,
      top: `${rect.top - 7}px`,
      width: `${rect.width + 7 * 2}px`,
      height: `${rect.height + 7 * 2}px`,
      zIndex: 1
    })
    
    this.moveRect = createSvg(
      'rect',
      {
        class: 'move',
        'data-id': 'move',
        stroke: '#1460F3',
        'stroke-width': 1,
        'stroke-dasharray': 2,
        'fill-opacity': 0,
        width: rect.width + 8,
        height: rect.height + 8,
        x: 3,
        y: 3
      }
    )
    
    this.topLeftRect = createSvg(
      'rect',
      {
        class: 'nw-resize',
        'data-id': 'nw-resize',
        fill: '#1460F3',
        stroke: 'none',
        width: 6,
        height: 6,
        x: 0,
        y: 0
      }
    )

    this.bottomLeftRect = createSvg(
      'rect',
      {
        class: 'sw-resize',
        'data-id': 'sw-resize',
        fill: '#1460F3',
        stroke: 'none',
        width: 6,
        height: 6,
        x: 0,
        y: rect.height + 8
      }
    )
    
    this.bottomRightRect = createSvg(
      'rect',
      {
        class: 'se-resize',
        'data-id': 'se-resize',
        fill: '#1460F3',
        stroke: 'none',
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: rect.height + 8
      }
    )

    this.topRightRect = createSvg(
      'rect',
      {
        class: 'ne-resize',
        'data-id': 'ne-resize',
        fill: '#1460F3',
        stroke: 'none',
        width: 6,
        height: 6,
        x: rect.width + 8,
        y: 0
      }
    )

    this.outerLine.append(this.moveRect)
    this.outerLine.append(this.topLeftRect)
    this.outerLine.append(this.bottomLeftRect)
    this.outerLine.append(this.bottomRightRect)
    this.outerLine.append(this.topRightRect)
    this.outerLineContainer.append(this.outerLine)

    this.initial = true

    if (this.show) {
      !this.layer.addMode && this.selectAnnotation()
      this.show = false
    }
  }

  selectAnnotation () {
    if (this.layer.annotationStore.selectedElementName === this.annotation.name) return
    this.handleClick()
  }

  handleClick (event) {
    if (!this.hidden || this.layer.annotationStore.creating || ['redaction', 'remove'].includes(this.layer.tool) || document.querySelector('.compare-document-container').contains(event?.target)) return
    this.hidden = false
    this.updateTool()

    if (this.annotation.measure) {
      this.layer.eventBus._on('measurePropertyPanelChanged', this.handlePropertyPanel)
      this.layer.eventBus._on('measurementPopupClicked', this.onHandlePopup)
    }
    document.addEventListener('keydown', this.onKeydown)

    const ignoreList = [
      this.svgElement,
      this.outerLine,
      document.querySelector('.toggle-button'),
      document.getElementById('measurePropertyPanelButton'),
      document.querySelector('.measure-panel'),
      document.querySelector('.measurement-board'),
      document.querySelector('.measurements-settings'),
      document.querySelector('.measurement-popup')
    ]
    this.deletetButton && ignoreList.push(this.deletetButton)
    this.replyButton && ignoreList.push(this.replyButton)
    onClickOutsideUp(ignoreList, this.handleOutside.bind(this))
  }

  handleOutside () {
    if (!this.hidden && this.annotation.measure && this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.eventBus.dispatch('changeRightPanelBtnDisabled')
      this.layer.eventBus.dispatch('measurePropertyChange', { isOpen: false })
      this.hidePopup()
    }

    this.hidden = true
    if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.annotationStore.selectedElementName = null
    }

    document.removeEventListener('keydown', this.onKeydown)
    this.outerLine?.removeEventListener('mousedown', this.onMousedown)
    this.outerLine?.removeEventListener('touchstart', this.onMousedown)
    this.outerLineContainer?.remove()

    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    scrollElement.style.touchAction = 'auto'

    if (this.annotation.measure) {
      this.layer.eventBus._off('measurePropertyPanelChanged', this.handlePropertyPanel)
      this.layer.eventBus._off('measurementPopupClicked', this.onHandlePopup)
    }
  }

  getActualRect (s) {
    const annotation = this.annotation
    const { left: x1, top: y1, right: x2, bottom: y2 } = annotation.rect
    
    const start = getActualPoint({ x: x1, y: y1 }, s)
    const end = getActualPoint({ x: x2, y: y2 }, s)
    
    return { start, end }
  }

  getInitialPoint () {
    const s = this.layer.viewport.scale
    const startX = this.start.x
    const startY = this.start.y
    const endX = this.end.x
    const endY = this.end.y

    let x1, y1, x2, y2
    x1 = startX / s
    y1 = startY / s
    x2 = endX / s
    y2 = endY / s
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    }
  }

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end })
    const actualbdwidth = (this.annotation.borderWidth || 2) * this.layer.viewport.scale
    let arrowLength = Math.max(actualbdwidth * 3, 10 * this.layer.viewport.scale)

    if (this.annotation.arrow) {
      return {
        top: initRect.top - arrowLength,
        left: initRect.left - arrowLength,
        width: initRect.width + arrowLength * 2,
        height: initRect.height + arrowLength * 2,
      }
    } else {
      return {
        top: initRect.top - actualbdwidth / 2,
        left: initRect.left - actualbdwidth / 2,
        width: initRect.width + actualbdwidth,
        height: initRect.height + actualbdwidth,
      }
    }
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    }
  }

  updateTool () {
    if (!this.initial) return
    if (this.hidden) {
      if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
        this.layer.annotationStore.selectedElementName = null
      }
      this.outerLineContainer.remove()
      document.removeEventListener('keydown', this.onKeydown)
    } else {
      if (this.layer.annotationStore.selectedElementName !== this.annotation.name) {
        const annotation = this.annotation
        this.layer.annotationStore.selectedElementName = annotation.name

        const annotationData = {
          ...annotation
        }
        delete annotationData.annotPtr
        delete annotationData.pagePtr
        delete annotationData.operate
        delete annotationData.targetPage
        delete annotationData.arrow

        this.layer.eventBus.dispatch('annotationSelected', annotationData)
      }
      this.layer.div.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)
      document.addEventListener('keydown', this.onKeydown)

      if (this.annotation.measure) {
        this.layer.eventBus.dispatch('changeRightPanelBtnDisabled')
        this.layer.eventBus.dispatch('measurePropertyChange', this.annotation)
        this.layer.eventBus.dispatch('updateMeasureData', {
          scale: this.annotation.scale,
          precision: this.annotation.precision,
          length: this.annotation.perimeter
        })
        this.showPopup()
      }

      const documentViewer = this.layer.documentViewer
      const scrollElement = documentViewer.getScrollViewElement()
      scrollElement.style.touchAction = 'none'
    }
  }

  handleMouseDown (event) {
    if (event.button !== 0 && event.type === 'mousedown') return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    const operatorId = event.target.getAttribute('data-id')
  
    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    this.scrollTop = scrollElement.scrollTop
    this.scrollLeft = scrollElement.scrollLeft

    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      operator: operatorId,
      clickX: pageX,
      clickY: pageY,
    }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  }

  handleMouseMove (event) {
    if (event.button !== 0 && event.type === 'mousemove') return
    this.hidePopup()
    
    if (event.type === 'touchmove') {
      this.layer.pageDiv.style.touchAction = 'none'
    }
    this.moving = true
    let { pageX, pageY } = getClickPoint(event)

    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    const scrollTop = scrollElement.scrollTop
    const scrollLeft = scrollElement.scrollLeft
    pageX = pageX + scrollLeft - this.scrollLeft
    pageY = pageY + scrollTop - this.scrollTop

    const start = this.start
    const end = this.end
    const startState = this.startState
    const { width, height } = this.layer.viewport

    if (startState.operator === 'nw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x)
      let y1 = pageY - (startState.clickY - this.start.y)
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(end.y - 5, y1)
      y1 = Math.max(0, y1)

      this.update({
        start: { x: x1, y: y1 },
        end,
      })
    } else if (startState.operator === 'sw-resize') {
      let x1 = pageX - (startState.clickX - this.start.x)
      let y1 = pageY - (startState.clickY - this.end.y)
      x1 = Math.min(end.x - 5, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(height, y1)
      y1 = Math.max(start.y + 5, y1)

      this.update({
        start: { x: x1, y: start.y },
        end: { x: end.x, y: y1 },
      })
    } else if (startState.operator === 'se-resize') {
      let x = pageX - (startState.clickX - this.end.x)
      let y = pageY - (startState.clickY - this.end.y)
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(height, y)
      y = Math.max(start.y + 5, y)
      this.update({
        start,
        end: { x, y },
      })
    } else if (startState.operator === 'ne-resize') {
      let x = pageX - (startState.clickX - this.end.x)
      let y = pageY - (startState.clickY - this.start.y)
      x = Math.min(width, x)
      x = Math.max(start.x + 5, x)
      y = Math.min(end.y - 5, y)
      y = Math.max(0, y)
      this.update({
        start: { x: start.x, y },
        end: { x, y: end.y },
      })
    } else if (startState.operator === 'move') {
      let x1, y1, x2, y2
      const documentViewer = this.layer.documentViewer

      let windowCoordinates = this.getMouseLocation(event)
      const page = documentViewer.getSelectedPage(windowCoordinates, windowCoordinates)
      if (page.first && page.first !== (this.layer.page + 1)) {
        const pagePoint = documentViewer.windowToPage(windowCoordinates, page.first, false)

        const annotationWidth = Math.abs(start.x - end.x)
        const annotationHeight = Math.abs(start.y - end.y)
        if (documentViewer.pdfViewer.spreadMode === 0 && documentViewer.pdfViewer.scrollMode === 0) {
          x1 = pageX - (startState.clickX - this.start.x)
          y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
          x2 = pageX - (startState.clickX - this.end.x)
          y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
        } else if (documentViewer.pdfViewer.spreadMode === 0 && documentViewer.pdfViewer.scrollMode === 1) {
          x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
          y1 = pageY - (startState.clickY - this.start.y)
          x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
          y2 = pageY - (startState.clickY - this.end.y)
        } else if (documentViewer.pdfViewer.spreadMode === 1 && documentViewer.pdfViewer.scrollMode === 0) {
          const minPage = Math.min(page.first, this.layer.page + 1)
          const maxPage = Math.max(page.first, this.layer.page + 1)
          if ((maxPage - minPage) > 2 || ((minPage + 1) === maxPage && minPage % 2 === 0)) {
            return
          }
          if ((minPage + 1) === maxPage) {
            x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
            y1 = pageY - (startState.clickY - this.start.y)
            x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
            y2 = pageY - (startState.clickY - this.end.y)
          } else if ((minPage + 2) === maxPage) {
            x1 = pageX - (startState.clickX - this.start.x)
            y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
            x2 = pageX - (startState.clickX - this.end.x)
            y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
          }
        } else if (documentViewer.pdfViewer.spreadMode === 2 && documentViewer.pdfViewer.scrollMode === 0) {
          const minPage = Math.min(page.first, this.layer.page + 1)
          const maxPage = Math.max(page.first, this.layer.page + 1)
          if ((maxPage - minPage) > 2 || ((minPage + 1) === maxPage && !(minPage === 1 && maxPage === 2) && maxPage % 2 === 0 || (minPage === 1 && maxPage === 3))) {
            return
          }
          if ((minPage + 2) === maxPage || (minPage === 1 && maxPage === 2)) {
            x1 = pageX - (startState.clickX - this.start.x)
            y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
            x2 = pageX - (startState.clickX - this.end.x)
            y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
          } else if ((minPage + 1) === maxPage) {
            x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
            y1 = pageY - (startState.clickY - this.start.y)
            x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
            y2 = pageY - (startState.clickY - this.end.y)
          }
        }
        this.layer = documentViewer.pdfViewer._pages[page.first - 1].compdfAnnotationLayer

        this.annotationContainer.remove()
        this.outerLineContainer.remove()
        const container = this.layer.div
        container.append(this.annotationContainer)
        container.append(this.outerLineContainer)
      } else {
        y1 = pageY - (startState.clickY - this.start.y)
        x1 = pageX - (startState.clickX - this.start.x)
        x2 = pageX - (startState.clickX - this.end.x)
        y2 = pageY - (startState.clickY - this.end.y)
        const windowPoint1 = documentViewer.pageToWindow({
          x: x1,
          y: y1
        }, this.layer.page + 1, false)
        const windowPoint2 = documentViewer.pageToWindow({
          x: x2,
          y: y2
        }, this.layer.page + 1, false)
        let x = 0, y = 0
        if (pageY > startState.clickY) {
          y = scrollElement.offsetHeight - (windowPoint2.y - scrollElement.scrollTop - scrollElement.offsetTop)
          y = y < 100 ? 5 : 0
        } else if (pageY < startState.clickY) {
          y = windowPoint1.y - scrollElement.scrollTop - scrollElement.offsetTop
          y = y < 100 ? -5 : 0
        }
        if (pageX > startState.clickX) {
          x = scrollElement.offsetWidth - (windowPoint2.x - scrollElement.scrollLeft - scrollElement.offsetLeft)
          x = x < 100 ? 5 : 0
        } else if (pageX < startState.clickX) {
          x = (windowPoint1.x - scrollElement.scrollLeft - scrollElement.offsetLeft)
          x = x < 100 ? -5 : 0
        }

        if (x || y) {
          scrollElement.scrollTop += y
          scrollElement.scrollLeft += x
        }
        const rect = {
          width: Math.abs(start.x - end.x),
          height: Math.abs(start.y - end.y)
        }
        if (x1 < x2) {
          x1 = Math.max(0, x1)
          x1 = Math.min(width - rect.width, x1)
          x2 = Math.max(rect.width, x2)
          x2 = Math.min(width, x2)
        } else {
          x2 = Math.max(0, x2)
          x2 = Math.min(width - rect.width, x2)
          x1 = Math.max(rect.width, x1)
          x1 = Math.min(width, x1)
        }

        if (y1 < y2) {
          y1 = Math.max(0, y1)
          y1 = Math.min(height - rect.height, y1)
          y2 = Math.max(rect.height, y2)
          y2 = Math.min(height, y2)
        } else {
          y2 = Math.max(0, y2)
          y2 = Math.min(height - rect.height, y2)
          y1 = Math.max(rect.height, y1)
          y1 = Math.min(height, y1)
        }
      }

      this.update({
        start: { x: x1, y: y1 },
        end: { x: x2, y: y2 },
      });
    }
    startState.clickX = pageX
    startState.clickY = pageY
  }

  handleDelete (event) {
    if (this.layer.tool && event) {
      event.stopPropagation()
    }

    this.delete()
    const annotationData = {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        pageIndex: this.layer.page
      }
    }
    this.annotation.annotPtr && (annotationData.annotation.annotPtr = this.annotation.annotPtr)

    this.layer.eventBus.dispatch('annotationChanged', annotationData)
  }

  delete () {
    this.handleOutside()
    this.annotationContainer?.remove()
    this.annotation.isDelete = true
    this.layer.deleteAnnotationInLayer({
      name: this.annotation.name
    })
  }

  async handleMouseUp (event) {
    if (event.button !== 0 && event.type === 'mouseup') return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      this.layer.pageDiv.style.touchAction = ''
    }
    this.moving = false
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
    this.scrollLeft = 0
    this.scrollTop = 0

    const annotation = this.annotation
    this.annotation.measure && this.showPopup()

    if (annotation.pageIndex !== this.layer.page) {
      const annotationData = {
        type: 'delete',
        annotation: {
          operate: "del-annot",
          name: annotation.name,
          pageIndex: annotation.pageIndex,
          annotPtr: annotation.annotPtr
        }
      }
      this.layer.eventBus.dispatch('annotationChanged', annotationData)

      const newAnnotationData = Object.assign({}, annotation)
      delete newAnnotationData.annotPtr
      delete newAnnotationData.pagePtr
      delete newAnnotationData.name
      newAnnotationData.operate = 'add-annot'
      newAnnotationData.pageIndex = this.layer.page
      // this.layer.eventBus.dispatch('annotationChanged', {
      //   type: 'add',
      //   show: true,
      //   annotation: newAnnotationData
      // })
      const annots = await this.layer.documentViewer.annotationManager.createMeasurementAnnotation(newAnnotationData)
      this.annotation = annots[0]
      await this.layer.documentViewer.annotationManager.drawAnnotationsFromList(this.annotation)

      this.annotationContainer.remove()
      this.outerLineContainer.remove()
    } else {
      const annotationData = {
        operate: "mod-annot",
        name: annotation.name,
        pageIndex: this.layer.page,
        pagePtr: annotation.pagePtr,
        annotPtr: annotation.annotPtr,
        rect: annotation.rect,
        linePoints: annotation.linePoints
      }
      // this.layer.eventBus.dispatch('annotationChanged', {
      //   type: 'modify',
      //   annotation: annotationData
      // })
      await this.layer.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
    }
  }

  openReply () {
    this.layer.eventBus.dispatch('openAnnotationReply', this.annotation)
  }

  /**
   * 更新绘制注释以及选中框的位置
   * @param {object} - 新的缩放后的对角坐标
   */
  update ({ start, end }) {
    this.start = start
    this.end = end

    const { start: initStart, end: initEnd } = this.getInitialPoint()
    this.annotation.linePoints = [initStart.x, initEnd.y, initEnd.x, initStart.y]
    this.updateRect()
    const rect = this.rect
    const actualbdwidth = (this.annotation.borderWidth || 2) * this.layer.viewport.scale

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'

    this.svgElement.setAttribute('width', rect.width)
    this.svgElement.setAttribute('height', rect.height)
    this.svgElement.setAttribute('viewBox', `0 0 ${rect.width} ${rect.height}`)

    this.shapeEle.setAttribute('cx', rect.width / 2)
    this.shapeEle.setAttribute('cy', rect.height / 2)
    this.shapeEle.setAttribute('rx', (rect.width - actualbdwidth) / 2)
    this.shapeEle.setAttribute('ry', (rect.height - actualbdwidth) / 2)

    this.outerLine.setAttribute('width', rect.width + 7 * 2)
    this.outerLine.setAttribute('height', rect.height + 7 * 2)
    this.outerLine.setAttribute('viewBox', `0 0 ${rect.width + 7 * 2} ${rect.height + 7 * 2}`)
    this.outerLine.style.left = `${rect.left - 7}px`
    this.outerLine.style.top = `${rect.top - 7}px`
    this.outerLine.style.width = `${rect.width + 7 * 2}px`
    this.outerLine.style.height = `${rect.height + 7 * 2}px`

    if (this.deletetButton) {
      this.deletetButton.style.left = (rect.left + rect.width + MARGIN_DISTANCE - 30) + 'px'
      this.deletetButton.style.top = (rect.top - MARGIN_DISTANCE - 30) + 'px'
    }
    if (this.replyButton) {
      this.replyButton.style.left = (rect.left + rect.width + MARGIN_DISTANCE - 65) + 'px'
      this.replyButton.style.top = (rect.top - MARGIN_DISTANCE - 30) + 'px'
    }

    this.moveRect.setAttribute('width',  rect.width + 8)
    this.moveRect.setAttribute('height',  rect.height + 8)

    this.bottomLeftRect.setAttribute('y', rect.height + 8)
    this.bottomRightRect.setAttribute('x', rect.width + 8)
    this.bottomRightRect.setAttribute('y', rect.height + 8)
    this.topRightRect.setAttribute('x', rect.width + 8)

    this.textEle.setAttribute('x', rect.width / 2 + this.annotation.borderWidth)
    this.textEle.setAttribute('y', rect.height / 2 + this.annotation.borderWidth + this.annotation.fontSize)
    this.getResult()
    this.textEle.textContent = this.annotation.contents

    // this.annotation.measure && isUpdateMeasure && this.updateMeasure()
  }

  async updateMeasure () {
    // 更新测量数据
    if (this.startState && this.startState.operator !== 'move') {
      const annotationData = {
        pagePtr: this.annotation.pagePtr,
        annotPtr: this.annotation.annotPtr,
        rulerPtr: this.annotation.rulerPtr,
        pageIndex: this.annotation.pageIndex,
        rect: this.annotation.rect,
        linePoints: this.annotation.linePoints,
      }
      if (this.annotation.contents) {
        annotationData.contents = this.annotation.contents
        annotationData.labelPosition = this.annotation.labelPosition
      }

      await this.layer.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
      this.getResult()
    }

    if (this.startState && this.startState.operator === 'move') return

    // 更新看板数据
    this.layer.eventBus.dispatch('updateMeasureData', {
      scale: this.annotation.scale,
      precision: this.annotation.precision,
      length: this.annotation.perimeter
    })
  }

  async handlePropertyPanel(props) {
    const annotation = this.annotation
    if (this.layer.annotationStore.selectedElementName !== annotation.name) return

    let updatePropsNum = 0

    for (const key in props) {
      updatePropsNum += await this.setCustomData(key, props[key], true)
    }

    updatePropsNum > 0 && this.sendData(props)
  }

  async setCustomData(key, value, noSend = false) {
    const annotation = this.annotation
    let isUpdate = 0

    for (const item in annotation) {
      if (item === key && this.annotation[item] !== value) {
        annotation[item] = value
        isUpdate = 1

        // Styles
        if (key === 'borderColor') {
          const lineNames = ['startLine', 'endLine', 'startLead', 'endLead', 'startLeadExtension', 'endLeadExtension']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            line && line.setAttribute('stroke', value)
          })
        }
        if (key === 'opacity') {
          this.svgElement.style.opacity = value
        }
        if (key === 'borderWidth') {
          const lineNames = ['startLine', 'endLine', 'startLead', 'endLead', 'startLeadExtension', 'endLeadExtension']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            line && line.setAttribute('stroke-width', value)
          })
          this.update({ start: this.start, end: this.end }, false)
          this.showPopup()
        }
        if (key === 'borderStyle') {
          const lineNames = ['startLine', 'endLine']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            if (line && value === 'solid') {
              line.setAttribute('stroke-dasharray', '')
            }
            if (line && value === 'dashed') {
              let dashString = ''
              const scaleValue = Math.ceil(this.ratio * this.layer.viewport.scale)

              if (annotation.dashes.length === 1 || (annotation.dashes.length === 2 && annotation.dashes[0] === annotation.dashes[1])) {
                dashString = Math.ceil(annotation.dashes[0] * scaleValue) + ''
              } else if (annotation.dashes.length > 2) {
                dashString = annotation.dashes.map(dash => Math.ceil(dash * scaleValue)).join(', ')
              }
              line.setAttribute('stroke-dasharray', dashString)
            }
          })
        }
        if (key === 'dashes') {
          const lineNames = ['startLine', 'endLine']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            line && line.setAttribute('stroke-dasharray', Math.ceil(value[0] * this.ratio * this.layer.viewport.scale))
          })
        }
        if (key === 'color') {
          this.textEle.setAttribute('fill', value)
        }
        if (key === 'fontSize') {
          this.textEle.setAttribute('font-size', value)
        }
        if (key === 'fontName') {
          this.textEle.style.fontFamily = value
        }
        if (key === 'fontStyle') {
          const fontStyle = value
          if (fontStyle === 'boldItalic') {
            this.textEle.style.fontStyle = 'italic'
            this.textEle.style.fontWeight = 'bold'
            annotation.isBold = true
            annotation.isItalic = true
          } else if (fontStyle === 'bold') {
            this.textEle.style.fontStyle = 'normal'
            this.textEle.style.fontWeight = 'bold'
            annotation.isBold = true
            annotation.isItalic = false
          } else if (fontStyle === 'italic') {
            this.textEle.style.fontStyle = 'italic'
            this.textEle.style.fontWeight = 'normal'
            annotation.isBold = false
            annotation.isItalic = true
          } else {
            annotation.isItalic = false
            annotation.isBold = false
            this.textEle.style.fontStyle = 'normal'
            this.textEle.style.fontWeight = 'normal'
          }
        }
        if (key === 'contents') {
          this.textEle.textContent = value
        }

        // Measure Properties
        if (key === 'scale' || key === 'precision') {
          if (key === 'precision') {
            this.decimalPlaces = Math.abs(annotation.precision).toString().split('.')[1]?.length || 0
          }

          await this.layer.documentViewer.annotationManager.editMeasurementAnnotation({
            rulerPtr: annotation.rulerPtr,
            [key]: value
          })
          this.getResult()
          await this.layer.documentViewer.annotationManager.editMeasurementAnnotation({
            annotPtr: annotation.annotPtr,
            contents: annotation.contents
          })
          
          const factor = await this.layer.documentViewer.messageHandler.sendWithPromise('GetFactor', annotation.rulerPtr)
          annotation.factor = factor

          if (this.textEle) {
            this.textEle.textContent = annotation.contents
          }
          if (this.layer.annotationStore.selectedElementName === this.annotation.name && !document.querySelector('.measure-panel').classList.contains('closed')) {
            this.layer.eventBus.dispatch('measurePropertyChange', this.annotation)
          }

          this.layer.eventBus.dispatch('updateMeasureData', {
            scale: value,
            length: annotation.perimeter
          })

          isUpdate = 0
        }

        if (['fontSize', 'scale', 'precision', 'contents'].includes(key)) {
          const startLine = this.svgElement.querySelector('[startLine]')
          const endLine = this.svgElement.querySelector('[endLine]')
          const dx = this.end.x - this.start.x
          const dy = this.end.y - this.start.y
          const length = Math.sqrt(dx * dx + dy * dy)
          const unitX = dx / length
          const unitY = dy / length
          const textGap = this.getTextWidth()

          if (startLine) {
            const line1End = {
              x: this.start.x + unitX * (length / 2 - textGap / 2),
              y: this.start.y + unitY * (length / 2 - textGap / 2),
            }
            startLine.setAttribute("x1", this.start.x)
            startLine.setAttribute("y1", this.start.y)
            startLine.setAttribute("x2", line1End.x)
            startLine.setAttribute("y2", line1End.y)
          }
          if (endLine) {
            const line2Start = {
              x: this.start.x + unitX * (length / 2 + textGap / 2),
              y: this.start.y + unitY * (length / 2 + textGap / 2),
            }
            endLine.setAttribute("x1", line2Start.x)
            endLine.setAttribute("y1", line2Start.y)
            endLine.setAttribute("x2", this.end.x)
            endLine.setAttribute("y2", this.end.y)
          }
        }
      }
    }
    !noSend && this.sendData({ [key]: value })
    return isUpdate
  }

  sendData(props) {
    const annotation = this.annotation
    const annotationData = {
      pagePtr: annotation.pagePtr,
      annotPtr: annotation.annotPtr,
      rulerPtr: annotation.rulerPtr,
      pageIndex: annotation.pageIndex,
      contents: annotation.contents
    }

    const isFont = ['fontSize', 'fontName', 'color', 'fontStyle'].some(item => props[item])
    if (isFont) {
      annotationData.fontSize = annotation.fontSize
      annotationData.color = annotation.color
      const fontArr = {
        Courier: {
          normal: 'Courier',
          bold: 'Courier-Bold',
          italic: 'Courier-Oblique',
          boldItalic: 'Courier-BoldOblique'
        },
        Helvetica: {
          normal: 'Helvetica',
          bold: 'Helvetica-Bold',
          italic: 'Helvetica-Oblique',
          boldItalic: 'Helvetica-BoldOblique'
        },
        Times: {
          normal: 'Times',
          bold: 'Times-Bold',
          italic: 'Times-Italic',
          boldItalic: 'Times-BoldItalic'
        }
      }

      const fontName = annotation.fontName
      if (annotation.isBold && annotation.isItalic) {
        annotationData.fontName = fontArr[fontName].boldItalic
      } else if (annotation.isBold) {
        annotationData.fontName = fontArr[fontName].bold
      } else if (annotation.isItalic) {
        annotationData.fontName = fontArr[fontName].italic
      } else {
        annotationData.fontName = fontArr[fontName] ? fontArr[fontName].normal : fontName
      }
      delete props.fontName
    }
    if (props.dashes) {
      annotationData.borderStyle = annotation.borderStyle
      annotationData.borderWidth = annotation.borderWidth
    }
    if (props.borderStyle) {
      annotationData.dashes = props.borderStyle === 'solid' ? [] : annotation.dashes
      annotationData.borderWidth = annotation.borderWidth
    }

    const updateAnnot = Object.assign({}, annotationData, props)
    this.layer.documentViewer.annotationManager.editMeasurementAnnotation(updateAnnot)
  }

  /** 计算测量结果，并更新至注释content属性上 */
  getResult() {
    // const { annotPtr, rulerPtr } = this.annotation
    // const value = await this.layer.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { annotPtr, rulerPtr })
    // const formatValue = this.formatValue(value)

    const { rect, factor, scale: { translateUnit } } = this.annotation

    const width = Math.abs(rect.right - rect.left)
    const height = Math.abs(rect.bottom - rect.top)
    const a = Math.max(width, height) / 2
    const b = Math.min(width, height) / 2

    const area = this.formatValue(Math.PI * a * b * factor ** 2)
    const perimeter = this.formatValue(Math.PI * (3 * (a + b) - Math.sqrt((3 * a + b) * (a + 3 * b))) * factor)

    this.annotation.area = area
    this.annotation.perimeter = perimeter
    this.annotation.contents = area + translateUnit + '\u00B2' + '\n' + perimeter + translateUnit
  }

  showPopup() {
    const pageRect = this.layer.div.getBoundingClientRect()

    const left = pageRect.left + this.rect.left
    const top = this.rect.top + pageRect.top
    const right = left + this.rect.width + 8
    const bottom = top + this.rect.height + 8
    const rect = {
      left,
      top,
      right,
      bottom
    }
    this.layer.eventBus.dispatch('showMeasurementPopup', { rect })
    this.isPopupShow = true
  }

  hidePopup() {
    if (!this.isPopupShow) return
    this.isPopupShow = false
    this.layer.eventBus.dispatch('showMeasurementPopup')
  }

  async handlePopup(data) {
    if (data === 'delete') {
      this.handleDelete()
    }
    if (data === 'reply') {
      this.openReply()
    }
  }

  // 根据位数保留小数
  formatValue(value) {
    return value % 1 === 0 ? value : Number(value.toFixed(this.decimalPlaces))
  }

  /**
   * 获取文本宽度
   * @param {string} text - 文本
   * @param {number} lineLength - 文本所在线段的长度
   */
  getTextWidth(text, lineLength = 0) {
    const canvas = document.createElement('canvas')
    const context = canvas.getContext('2d')

    context.font = `${this.annotation.fontSize}px ${this.annotation.fontFamily}`
    const metrics = context.measureText(text)
    const gap = metrics.width * 1.1

    return lineLength > gap ? gap : lineLength
  }

  async handleKeyDown (e) {
    if (e.target.tagName === 'INPUT' || e.target.tagName === 'TEXTAREA') return
    
    let keyCode = e.keyCode || e.which
    if (keyCode === 8 || keyCode === 46) { // 8 backspace键，46 delete键
      this.handleDelete()
    }
  }

  /** 更新rect */
  updateRect() {
    const [ startX, endY, endX, startY ] = this.annotation.linePoints
    this.annotation.rect = {
      left: Math.min(startX, endX),
      top: Math.min(startY, endY),
      right: Math.max(startX, endX),
      bottom: Math.max(startY, endY)
    }
    const { start, end } = this.getActualRect(this.layer.viewport.scale)
    this.start = start
    this.end = end
    this.rect = this.calculate(start, end)
  }

  /** 绘制注释内容 */
  draw() {
    const annotation = this.annotation
    const rect = this.rect
    const scale = this.layer.viewport.scale
    const actualbdwidth = (annotation.borderWidth || 2) * scale

    let dashString = ''
    const scaleValue = Math.ceil(this.ratio * scale)
    if (annotation.dashes.length === 1 || (annotation.dashes.length === 2 && annotation.dashes[0] === annotation.dashes[1])) {
      dashString = Math.ceil(annotation.dashes[0] * scaleValue) + ''
    } else if (annotation.dashes.length > 2) {
      dashString = annotation.dashes.map(dash => Math.ceil(dash * scaleValue)).join(', ')
    }
    this.svgElement.innerHTML = ''

    this.shapeEle = createSvg('ellipse', {
      'cx': rect.width / 2,
      'cy': rect.height / 2,
      'rx': (rect.width - actualbdwidth) / 2,
      'ry': (rect.height - actualbdwidth) / 2,
      'stroke': annotation.borderColor,
      'stroke-dasharray': dashString,
      'stroke-width': actualbdwidth,
      'stroke-opacity': annotation.opacity || 1,
      'fill': annotation.fillColor ? annotation.fillColor : 'none',
      'fill-opacity': annotation.opacity || 1
    })
    this.svgElement.append(this.shapeEle)

    this.textEle = createSvg('text', {
      'fill': annotation.color || annotation.borderColor,
      'text-anchor': 'middle',
      'dominant-baseline': 'middle',
      'font-size': annotation.fontSize,
      'x': rect.width / 2,
      'y': rect.height / 2
    }, {
      pointerEvents: 'none'
    })
    this.textEle.textContent = annotation.contents
    this.svgElement.append(this.textEle)
  }

  /**
   * 添加点 
   * @param {number} x - 实际x坐标
   * @param {number} y - 实际y坐标
   */
  addPathPoint(x, y) {
    const annotation = this.annotation
    const scale = this.layer.viewport.scale
    const oriVertice = getInitialPoint({ x, y }, scale)
    annotation.linePoints.push(oriVertice)
    this.realLinePoints.push({ x, y })
    let [{ x: x1, y: y1 }, { x: x2, y: y2 }] = this.realLinePoints.slice(-2)

    if (![x1, y1, x2, y2].every(Number.isFinite)) return

    if (x1 === x2 && y1 === y2) {
      x2 += 1
      y2 += 1
    }

    const lineEle = createSvg('line', {
      'data-id': 'line' + (this.realLinePoints.length - 2),
      'stroke': annotation.borderColor,
      'stroke-width': 2 * scale,
      'stroke-opacity': 0.8,
      'x1': x1,
      'y1': y1,
      'x2': x2,
      'y2': y2
    }, {
      'stroke-linecap': 'round'
    })
    this.svgElement.append(lineEle)
  }

  /**
   * 修改指定点的坐标
   * @param {number} index - 指定点的序号，从0开始
   * @param {number} x - 实际x坐标
   * @param {number} y - 实际y坐标
   */
  setPathPoint(index, x, y) {
    if (index > this.realLinePoints.length - 1 || index < 0) return

    const annotation = this.annotation
    const scale = this.layer.viewport.scale
    const oriVertice = getInitialPoint({ x, y }, scale)

    annotation.linePoints[index].x = oriVertice.x
    annotation.linePoints[index].y = oriVertice.y
    this.realLinePoints[index].x = x
    this.realLinePoints[index].y = y

    if (index === 0) {
      this.updateLineElement(index)
    } else if (index === this.realLinePoints.length - 1) {
      this.updateLineElement(index - 1)
    } else {
      this.updateLineElement(index - 1)
      this.updateLineElement(index)
    }

    const [{ x: x1, y: y1 }, { x: x2, y: y2 }] = this.realLinePoints.slice(-2).map(p => p ?? { x: NaN, y: NaN })

    if (![x1, y1, x2, y2].every(Number.isFinite) || (x1 === x2 && y1 === y2)) return

    this.layer.eventBus.dispatch('updateMeasureData', {
      scale: this.annotation.scale,
      precision: this.annotation.precision,
      length: this.annotation.perimeter,
      angle: this.annotation.angle || 0
    })
  }

  /**
   * 更新线段的渲染
   * @param {number} index - 线段的序号（从0开始）
   */
  updateLineElement(index) {
    const { x: x1, y: y1 } = this.realLinePoints[index]
    const { x: x2, y: y2 } = this.realLinePoints[index + 1]

    if (![x1, y1, x2, y2].every(Number.isFinite) || (x1 === x2 && y1 === y2)) return
    
    const line = this.svgElement.querySelector('line[data-id="line' + index + '"]')

    line.setAttribute('x1', x1)
    line.setAttribute('y1', y1)
    line.setAttribute('x2', x2)
    line.setAttribute('y2', y2)
  }
}
