import Base from './base';
import { onClickOutsideUp } from '../ui_utils'
import { getClickPoint, createSvg, createElement, getActualPoint, getInitialPoint } from './utils';
import { MARGIN_DISTANCE } from '../../constants'

export default class Curve extends Base {
  constructor ({
    annotation,
    layer,
    show = false,
    enableReply = true
  }) {
    super({
      annotation,
      layer,
    })
    this.show = show
    this.enableReply = enableReply

    this.hidden = true
    this.initial = false

    this.start = null
    this.end = null
    this.rect = null
    this.realVertices = null

    this.svgElement = null

    this.ratio = window.devicePixelRatio || 1
    this.decimalPlaces = 0 // 保留小数位
    this.isPopupShow = false

    this.onClick = this.handleClick.bind(this)
    this.onMousedown = this.handleMouseDown.bind(this)
    this.onMouseup = this.handleMouseUp.bind(this)
    this.onMousemove = this.handleMouseMove.bind(this)
    this.onDelete = this.handleDelete.bind(this)
    this.onOpenReply = this.openReply.bind(this)
    this.handlePropertyPanel = this.handlePropertyPanel.bind(this)
    this.onHandlePopup = this.handlePopup.bind(this)
    this.onKeydown = this.handleKeyDown.bind(this)
    
    this.render()
  }

  async render () {
    const annotation = this.annotation
    const scale = this.layer.viewport.scale
    const vertices = this.estimateControlPoints(annotation.inkPointes[0])
    this.realVertices = vertices.map(item => getActualPoint(item, scale))
    this.updateRect()
    const rect = this.rect

    if (annotation.precision) {
      if (annotation.precision > 1) {
        annotation.precision = Math.pow(10, -Math.floor(Math.log10(annotation.precision)))
      }
      this.decimalPlaces = Math.abs(annotation.precision).toString().split('.')[1]?.length || 0
    }
    if (!annotation.borderColor) {
      annotation.borderColor = annotation.color || '#FF0000'
    }
    if (!annotation.factor) {
      annotation.factor = await this.layer.documentViewer.messageHandler.sendWithPromise('GetFactor', annotation.rulerPtr)
    }
    if (!annotation.fontStyle) {
      annotation.fontStyle = 'regular'
    }
    if (!annotation.contents) {
      this.getResult()
    }

    this.annotationContainer = createElement(
      'div',
      {
        top: rect.top + 'px',
        left: rect.left + 'px',
        width: rect.width + 'px',
        height: rect.height + 'px'
      }, {
        id: annotation.name,
        class: 'annotation'
      }
    )
    this.svgElement = createSvg(
      'svg',
      {
        viewBox: `${rect.left} ${rect.top} ${rect.width} ${rect.height}`
      }, {
        pointerEvents: 'all',
        overflow: 'visible',
        opacity: annotation.opacity ? annotation.opacity : ''
      }
    )
    this.annotationContainer.append(this.svgElement)
    this.svgElement.addEventListener('mousedown', this.onClick)
    this.layer.div.append(this.annotationContainer)
    this.draw()

    this.outerLineContainer = createElement('div', {}, { class: 'outline-container' })
    const outerLine = createSvg(
      'svg',
      {
        'data-id': 'move',
        width: rect.width,
        height: rect.height,
        viewBox: `${rect.left} ${rect.top} ${rect.width} ${rect.height}`
      }, {
        position: 'absolute',
        zIndex: 1,
        left: `${rect.left}px`,
        top: `${rect.top}px`,
        width: `${rect.width}px`,
        height: `${rect.height}px`,
        border: `${parseFloat(this.layer.viewport.scale.toFixed(1))}px dashed #1460F3`,
        boxSizing: 'content-box',
        pointerEvents: 'all',
        cursor: 'all-scroll',
        overflow: 'visible'
      }
    )
    this.moveLine = createSvg(
      'line',
      {
        class: 'move',
        'data-id': 'move',
        x1: rect.left,
        y1: rect.top,
        x2: rect.left + rect.width,
        y2: rect.top + rect.height,
        fill: 'transparent',
        stroke: 'transparent',
        'stroke-width': 5
      }
    )
    this.outerLine = outerLine
    this.outerLine.append(this.moveLine)
    this.outerLineContainer.append(this.outerLine)
    this.updateCircle()

    this.initial = true

    if (this.show) {
      !this.layer.addMode && this.selectAnnotation()
      this.show = false
    }
  }

  selectAnnotation () {
    if (this.layer.annotationStore.selectedElementName === this.annotation.name) return
    this.handleClick()
  }

  handleClick (event) {
    if (!this.hidden || this.layer.annotationStore.creating || ['redaction', 'remove'].includes(this.layer.tool) || document.querySelector('.compare-document-container').contains(event?.target)) return
    this.hidden = false
    this.updateTool()

    if (this.annotation.measure) {
      this.layer.eventBus._on('measurePropertyPanelChanged', this.handlePropertyPanel)
      this.layer.eventBus._on('measurementPopupClicked', this.onHandlePopup)
    }
    document.addEventListener('keydown', this.onKeydown)

    const ignoreList = [
      this.svgElement,
      this.outerLine,
      document.querySelector('.toggle-button'),
      document.getElementById('measurePropertyPanelButton'),
      document.querySelector('.measure-panel'),
      document.querySelector('.measurement-board'),
      document.querySelector('.measurements-settings'),
      document.querySelector('.measurement-popup')
    ]
    this.deletetButton && ignoreList.push(this.deletetButton)
    this.replyButton && ignoreList.push(this.replyButton)
    onClickOutsideUp(ignoreList, this.handleOutside.bind(this))
  }

  handleOutside () {
    if (!this.hidden && this.annotation.measure && this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.eventBus.dispatch('changeRightPanelBtnDisabled')
      this.layer.eventBus.dispatch('measurePropertyChange', { isOpen: false })
      this.hidePopup()
    }

    this.hidden = true
    if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
      this.layer.annotationStore.selectedElementName = null
    }

    document.removeEventListener('keydown', this.onKeydown)
    this.outerLine?.removeEventListener('mousedown', this.onMousedown)
    this.outerLine?.removeEventListener('touchstart', this.onMousedown)
    this.outerLineContainer?.remove()

    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    scrollElement.style.touchAction = 'auto'

    if (this.annotation.measure) {
      this.layer.eventBus._off('measurePropertyPanelChanged', this.handlePropertyPanel)
      this.layer.eventBus._off('measurementPopupClicked', this.onHandlePopup)
    }
  }

  getActualRect (s) {
    const annotation = this.annotation
    const { left: x1, top: y1, right: x2, bottom: y2 } = annotation.rect
    
    const start = getActualPoint({ x: x1, y: y1 }, s)
    const end = getActualPoint({ x: x2, y: y2 }, s)
    
    return { start, end }
  }

  getInitialPoint () {
    const s = this.layer.viewport.scale
    const startX = this.start.x
    const startY = this.start.y
    const endX = this.end.x
    const endY = this.end.y

    let x1, y1, x2, y2
    x1 = startX / s
    y1 = startY / s
    x2 = endX / s
    y2 = endY / s
    return {
      start: {
        x: x1,
        y: y1,
      },
      end: {
        x: x2,
        y: y2,
      },
    }
  }

  calculate (start, end) {
    const initRect = this.rectCalc({ start, end })
    const actualbdwidth = (this.annotation.borderWidth || 2) * this.layer.viewport.scale
    let arrowLength = Math.max(actualbdwidth * 3, 10 * this.layer.viewport.scale)

    if (this.annotation.arrow) {
      return {
        top: initRect.top - arrowLength,
        left: initRect.left - arrowLength,
        width: initRect.width + arrowLength * 2,
        height: initRect.height + arrowLength * 2,
      }
    } else {
      return {
        top: initRect.top - actualbdwidth / 2,
        left: initRect.left - actualbdwidth / 2,
        width: initRect.width + actualbdwidth,
        height: initRect.height + actualbdwidth,
      }
    }
  }

  rectCalc ({ start, end }) {
    return {
      top: start.y < end.y ? start.y : end.y,
      left: start.x < end.x ? start.x : end.x,
      width: Math.abs(end.x - start.x),
      height: Math.abs(end.y - start.y),
    }
  }

  /**
   * 通过顶点坐标计算注释rect
   * @param {Array<object>} vertices - 顶点坐标数组
   */
  rectCalcByVertices (vertices) {
    const arrayX = vertices.map(item => item.x)
    const arrayY = vertices.map(item => item.y)

    return {
      left: Math.min(...arrayX),
      top: Math.min(...arrayY),
      right: Math.max(...arrayX),
      bottom: Math.max(...arrayY)
    }
  }

  updateTool () {
    if (!this.initial) return
    if (this.hidden) {
      if (this.layer.annotationStore.selectedElementName === this.annotation.name) {
        this.layer.annotationStore.selectedElementName = null
      }
      this.outerLineContainer.remove()
      document.removeEventListener('keydown', this.onKeydown)
    } else {
      if (this.layer.annotationStore.selectedElementName !== this.annotation.name) {
        const annotation = this.annotation
        this.layer.annotationStore.selectedElementName = annotation.name

        const annotationData = {
          ...annotation
        }
        delete annotationData.annotPtr
        delete annotationData.pagePtr
        delete annotationData.operate
        delete annotationData.targetPage
        delete annotationData.arrow

        this.layer.eventBus.dispatch('annotationSelected', annotationData)
      }
      this.layer.div.append(this.outerLineContainer)
      this.outerLine.addEventListener('mousedown', this.onMousedown)
      this.outerLine.addEventListener('touchstart', this.onMousedown)
      document.addEventListener('keydown', this.onKeydown)

      if (this.annotation.measure) {
        this.layer.eventBus.dispatch('changeRightPanelBtnDisabled')
        this.layer.eventBus.dispatch('measurePropertyChange', this.annotation)
        this.layer.eventBus.dispatch('updateMeasureData', {
          scale: this.annotation.scale,
          precision: this.annotation.precision,
          length: this.annotation.perimeter
        })
        this.showPopup()
      }

      const documentViewer = this.layer.documentViewer
      const scrollElement = documentViewer.getScrollViewElement()
      scrollElement.style.touchAction = 'none'
    }
  }

  handleMouseDown (event) {
    if (event.button !== 0 && event.type === 'mousedown') return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    const operatorId = event.target.getAttribute('data-id')
  
    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    this.scrollTop = scrollElement.scrollTop
    this.scrollLeft = scrollElement.scrollLeft

    const { pageX, pageY } = getClickPoint(event)
    this.startState = {
      operator: operatorId,
      clickX: pageX,
      clickY: pageY,
    }

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    document.addEventListener('touchmove', this.onMousemove)
    document.addEventListener('touchend', this.onMouseup)
  }

  handleMouseMove (event) {
    if (event.button !== 0 && event.type === 'mousemove') return
    this.hidePopup()
    
    if (event.type === 'touchmove') {
      this.layer.pageDiv.style.touchAction = 'none'
    }
    this.moving = true
    let { pageX, pageY } = getClickPoint(event)

    const documentViewer = this.layer.documentViewer
    const scrollElement = documentViewer.getScrollViewElement()
    const scrollTop = scrollElement.scrollTop
    const scrollLeft = scrollElement.scrollLeft
    pageX = pageX + scrollLeft - this.scrollLeft
    pageY = pageY + scrollTop - this.scrollTop

    const start = this.start
    const end = this.end
    const startState = this.startState
    const { width, height, scale } = this.layer.viewport
    const regex = /^point(\d+)$/
    const match = regex.exec(startState.operator)

    if (match) {
      const index = +match[1]
      const point = this.realVertices[index]
      let x1 = pageX - (startState.clickX - point.x)
      let y1 = pageY - (startState.clickY - point.y)
      x1 = Math.min(width, x1)
      x1 = Math.max(0, x1)
      y1 = Math.min(height, y1)
      y1 = Math.max(0, y1)

      this.setPathPoint(index, x1, y1)
      this.update(index)
    } else if (startState.operator === 'move') {
      let x1, y1, x2, y2
      const documentViewer = this.layer.documentViewer

      let windowCoordinates = this.getMouseLocation(event)
      const page = documentViewer.getSelectedPage(windowCoordinates, windowCoordinates)
      if (page.first && page.first !== (this.layer.page + 1)) {
        const pagePoint = documentViewer.windowToPage(windowCoordinates, page.first, false)

        const annotationWidth = Math.abs(start.x - end.x)
        const annotationHeight = Math.abs(start.y - end.y)
        if (documentViewer.pdfViewer.spreadMode === 0 && documentViewer.pdfViewer.scrollMode === 0) {
          x1 = pageX - (startState.clickX - this.start.x)
          y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
          x2 = pageX - (startState.clickX - this.end.x)
          y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
        } else if (documentViewer.pdfViewer.spreadMode === 0 && documentViewer.pdfViewer.scrollMode === 1) {
          x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
          y1 = pageY - (startState.clickY - this.start.y)
          x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
          y2 = pageY - (startState.clickY - this.end.y)
        } else if (documentViewer.pdfViewer.spreadMode === 1 && documentViewer.pdfViewer.scrollMode === 0) {
          const minPage = Math.min(page.first, this.layer.page + 1)
          const maxPage = Math.max(page.first, this.layer.page + 1)
          if ((maxPage - minPage) > 2 || ((minPage + 1) === maxPage && minPage % 2 === 0)) {
            return
          }
          if ((minPage + 1) === maxPage) {
            x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
            y1 = pageY - (startState.clickY - this.start.y)
            x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
            y2 = pageY - (startState.clickY - this.end.y)
          } else if ((minPage + 2) === maxPage) {
            x1 = pageX - (startState.clickX - this.start.x)
            y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
            x2 = pageX - (startState.clickX - this.end.x)
            y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
          }
        } else if (documentViewer.pdfViewer.spreadMode === 2 && documentViewer.pdfViewer.scrollMode === 0) {
          const minPage = Math.min(page.first, this.layer.page + 1)
          const maxPage = Math.max(page.first, this.layer.page + 1)
          if ((maxPage - minPage) > 2 || ((minPage + 1) === maxPage && !(minPage === 1 && maxPage === 2) && maxPage % 2 === 0 || (minPage === 1 && maxPage === 3))) {
            return
          }
          if ((minPage + 2) === maxPage || (minPage === 1 && maxPage === 2)) {
            x1 = pageX - (startState.clickX - this.start.x)
            y1 = page.first > (this.layer.page + 1) ? pagePoint.y : pagePoint.y - annotationHeight
            x2 = pageX - (startState.clickX - this.end.x)
            y2 = page.first > (this.layer.page + 1) ? pagePoint.y + annotationHeight : pagePoint.y
          } else if ((minPage + 1) === maxPage) {
            x1 = page.first > (this.layer.page + 1) ? pagePoint.x : pagePoint.x - annotationWidth
            y1 = pageY - (startState.clickY - this.start.y)
            x2 = page.first > (this.layer.page + 1) ? pagePoint.x + annotationWidth : pagePoint.x
            y2 = pageY - (startState.clickY - this.end.y)
          }
        }
        this.layer = documentViewer.pdfViewer._pages[page.first - 1].compdfAnnotationLayer

        this.annotationContainer.remove()
        this.outerLineContainer.remove()
        const container = this.layer.div
        container.append(this.annotationContainer)
        container.append(this.outerLineContainer)
      } else {
        x1 = pageX - (startState.clickX - this.start.x);
        y1 = pageY - (startState.clickY - this.start.y);
        x2 = pageX - (startState.clickX - this.end.x);
        y2 = pageY - (startState.clickY - this.end.y);

        const windowPoint1 = documentViewer.pageToWindow({
          x: x1,
          y: y1
        }, this.layer.page + 1, false)
        const windowPoint2 = documentViewer.pageToWindow({
          x: x2,
          y: y2
        }, this.layer.page + 1, false)
        let x = 0, y = 0
        if (pageY > startState.clickY) {
          y = scrollElement.offsetHeight - (windowPoint2.y - scrollElement.scrollTop - scrollElement.offsetTop)
          y = y < 100 ? 5 : 0
        } else if (pageY < startState.clickY) {
          y = windowPoint1.y - scrollElement.scrollTop - scrollElement.offsetTop
          y = y < 100 ? -5 : 0
        }
        if (pageX > startState.clickX) {
          x = scrollElement.offsetWidth - (windowPoint2.x - scrollElement.scrollLeft - scrollElement.offsetLeft)
          x = x < 100 ? 5 : 0
        } else if (pageX < startState.clickX) {
          x = (windowPoint1.x - scrollElement.scrollLeft - scrollElement.offsetLeft)
          x = x < 100 ? -5 : 0
        }

        if (x || y) {
          scrollElement.scrollTop += y
          scrollElement.scrollLeft += x
        }

        if (x1 < x2) {
          x1 = Math.max(0, x1)
          x1 = Math.min(width - this.rect.width, x1)
          x2 = Math.max(this.rect.width, x2)
          x2 = Math.min(width, x2)
        } else {
          x2 = Math.max(0, x2)
          x2 = Math.min(width - this.rect.width, x2)
          x1 = Math.max(this.rect.width, x1)
          x1 = Math.min(width, x1)
        }

        if (y1 < y2) {
          y1 = Math.max(0, y1)
          y1 = Math.min(height - this.rect.height, y1)
          y2 = Math.max(this.rect.height, y2)
          y2 = Math.min(height, y2)
        } else {
          y2 = Math.max(0, y2)
          y2 = Math.min(height - this.rect.height, y2)
          y1 = Math.max(this.rect.height, y1)
          y1 = Math.min(height, y1)
        }
      }

      const deltaX = x1 - start.x
      const deltaY = y1 - start.y

      for (let i = 0; i < this.realVertices.length; i++) {
        const point = this.realVertices[i]
        const x = point.x + deltaX
        const y = point.y + deltaY
        
        this.setPathPoint(i, x, y)
      }

      this.update()
    }
    startState.clickX = pageX
    startState.clickY = pageY
  }

  handleDelete (event) {
    if (this.layer.tool && event) {
      event.stopPropagation()
    }

    this.delete()
    const annotationData = {
      type: 'delete',
      annotation: {
        operate: "del-annot",
        name: this.annotation.name,
        pageIndex: this.layer.page
      }
    }
    this.annotation.annotPtr && (annotationData.annotation.annotPtr = this.annotation.annotPtr)

    this.layer.eventBus.dispatch('annotationChanged', annotationData)
  }

  delete () {
    this.handleOutside()
    this.annotationContainer?.remove()
    this.annotation.isDelete = true
    this.layer.deleteAnnotationInLayer({
      name: this.annotation.name
    })
  }

  async handleMouseUp (event) {
    if (event.button !== 0 && event.type === 'mouseup') return
    if (this.layer.tool) {
      event.stopPropagation()
    }
    if (event.type === 'touchend') {
      this.layer.pageDiv.style.touchAction = ''
    }
    this.moving = false
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    document.removeEventListener('touchmove', this.onMousemove)
    document.removeEventListener('touchend', this.onMouseup)
    this.scrollLeft = 0
    this.scrollTop = 0

    const annotation = this.annotation
    this.annotation.measure && this.showPopup()

    if (annotation.pageIndex !== this.layer.page) {
      const annotationData = {
        type: 'delete',
        annotation: {
          operate: "del-annot",
          name: annotation.name,
          pageIndex: annotation.pageIndex,
          annotPtr: annotation.annotPtr
        }
      }
      this.layer.eventBus.dispatch('annotationChanged', annotationData)

      const newAnnotationData = Object.assign({}, annotation)
      delete newAnnotationData.annotPtr
      delete newAnnotationData.pagePtr
      delete newAnnotationData.name
      newAnnotationData.operate = 'add-annot'
      newAnnotationData.pageIndex = this.layer.page
      // this.layer.eventBus.dispatch('annotationChanged', {
      //   type: 'add',
      //   show: true,
      //   annotation: newAnnotationData
      // })
      const annots = await this.layer.documentViewer.annotationManager.createMeasurementAnnotation(newAnnotationData)
      this.annotation = annots[0]
      await this.layer.documentViewer.annotationManager.drawAnnotationsFromList(this.annotation)

      this.annotationContainer.remove()
      this.outerLineContainer.remove()
    } else {
      const annotationData = {
        operate: "mod-annot",
        name: annotation.name,
        pageIndex: this.layer.page,
        pagePtr: annotation.pagePtr,
        annotPtr: annotation.annotPtr,
        rect: annotation.rect,
        inkPointes: annotation.inkPointes
      }
      // this.layer.eventBus.dispatch('annotationChanged', {
      //   type: 'modify',
      //   annotation: annotationData
      // })
      await this.layer.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
    }
  }

  openReply () {
    this.layer.eventBus.dispatch('openAnnotationReply', this.annotation)
  }

  /**
   * 更新绘制注释以及选中框的位置
   * @param {number} index - 指定更新选中框中某个顶点按钮的坐标，若不指定则重绘所有顶点按钮
   */
  update (index) {
    this.updateRect()
    this.calculateInkPoints()
    const rect = this.rect

    this.annotationContainer.style.top = rect.top + 'px'
    this.annotationContainer.style.left = rect.left + 'px'
    this.annotationContainer.style.width = rect.width + 'px'
    this.annotationContainer.style.height = rect.height + 'px'

    this.svgElement.setAttribute("width", rect.width)
    this.svgElement.setAttribute("height", rect.height)
    this.svgElement.setAttribute("viewBox", `${rect.left} ${rect.top} ${rect.width} ${rect.height}`)

    this.outerLine.setAttribute("height", rect.height)
    this.outerLine.setAttribute("viewBox", `${rect.left} ${rect.top} ${rect.width} ${rect.height}`)
    this.outerLine.style.left = `${rect.left}px`
    this.outerLine.style.top = `${rect.top}px`
    this.outerLine.style.width = `${rect.width}px`
    this.outerLine.style.height = `${rect.height}px`

    if (this.deletetButton) {
      this.deletetButton.style.left = (rect.left + rect.width + MARGIN_DISTANCE - 30) + 'px'
      this.deletetButton.style.top = (rect.top - MARGIN_DISTANCE - 30) + 'px'
    }
    if (this.replyButton) {
      this.replyButton.style.left = (rect.left + rect.width + MARGIN_DISTANCE - 65) + 'px'
      this.replyButton.style.top = (rect.top - MARGIN_DISTANCE - 30) + 'px'
    }
    
    this.moveLine.setAttribute("x1", rect.left)
    this.moveLine.setAttribute("y1", rect.top)
    this.moveLine.setAttribute("x2", rect.left + rect.width)
    this.moveLine.setAttribute("y2", rect.top + rect.height)

    this.updateCircle(index ?? null)

    // this.annotation.measure && isUpdateMeasure && this.updateMeasure(start, end)
  }

  async updateMeasure (start, end) {
    // 更新测量数据
    if (this.startState && this.startState.operator !== 'move') {
      const initPoint = this.getInitialPoint()

      const annotationData = {
        pagePtr: this.annotation.pagePtr,
        annotPtr: this.annotation.annotPtr,
        rulerPtr: this.annotation.rulerPtr,
        pageIndex: this.annotation.pageIndex,
        rect,
        linePoints,
      }
      if (this.annotation.contents) {
        annotationData.contents = this.annotation.contents
        annotationData.labelPosition = this.annotation.labelPosition
      }
      this.annotation.rect = rect
      this.annotation.linePoints = linePoints

      await this.layer.documentViewer.annotationManager.editMeasurementAnnotation(annotationData)
      this.getResult()
    }

    if (this.startState && this.startState.operator === 'move') return

    // 更新看板数据
    this.layer.eventBus.dispatch('updateMeasureData', {
      scale: this.annotation.scale,
      precision: this.annotation.precision,
      length: this.annotation.perimeter
    })
  }

  async handlePropertyPanel(props) {
    const annotation = this.annotation
    if (this.layer.annotationStore.selectedElementName !== annotation.name) return

    let updatePropsNum = 0

    for (const key in props) {
      updatePropsNum += await this.setCustomData(key, props[key], true)
    }

    updatePropsNum > 0 && this.sendData(props)
  }

  async setCustomData(key, value, noSend = false) {
    const annotation = this.annotation
    let isUpdate = 0

    for (const item in annotation) {
      if (item === key && this.annotation[item] !== value) {
        annotation[item] = value
        isUpdate = 1

        // Styles
        if (key === 'borderColor') {
          const lineNames = ['startLine', 'endLine', 'startLead', 'endLead', 'startLeadExtension', 'endLeadExtension']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            line && line.setAttribute('stroke', value)
          })
        }
        if (key === 'opacity') {
          this.svgElement.style.opacity = value
        }
        if (key === 'borderWidth') {
          const lineNames = ['startLine', 'endLine', 'startLead', 'endLead', 'startLeadExtension', 'endLeadExtension']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            line && line.setAttribute('stroke-width', value)
          })
          this.update({ start: this.start, end: this.end }, false)
          this.showPopup()
        }
        if (key === 'borderStyle') {
          const lineNames = ['startLine', 'endLine']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            if (line && value === 'solid') {
              line.setAttribute('stroke-dasharray', '')
            }
            if (line && value === 'dashed') {
              let dashString = ''
              const scaleValue = Math.ceil(this.ratio * this.layer.viewport.scale)

              if (annotation.dashes.length === 1 || (annotation.dashes.length === 2 && annotation.dashes[0] === annotation.dashes[1])) {
                dashString = Math.ceil(annotation.dashes[0] * scaleValue) + ''
              } else if (annotation.dashes.length > 2) {
                dashString = annotation.dashes.map(dash => Math.ceil(dash * scaleValue)).join(', ')
              }
              line.setAttribute('stroke-dasharray', dashString)
            }
          })
        }
        if (key === 'dashes') {
          const lineNames = ['startLine', 'endLine']
          lineNames.forEach(name => {
            const line = this.svgElement.querySelector(`[${name}]`)
            line && line.setAttribute('stroke-dasharray', Math.ceil(value[0] * this.ratio * this.layer.viewport.scale))
          })
        }
        if (key === 'color') {
          this.textEle.setAttribute('fill', value)
        }
        if (key === 'fontSize') {
          this.textEle.setAttribute('font-size', value)
        }
        if (key === 'fontName') {
          this.textEle.style.fontFamily = value
        }
        if (key === 'fontStyle') {
          const fontStyle = value
          if (fontStyle === 'boldItalic') {
            this.textEle.style.fontStyle = 'italic'
            this.textEle.style.fontWeight = 'bold'
            annotation.isBold = true
            annotation.isItalic = true
          } else if (fontStyle === 'bold') {
            this.textEle.style.fontStyle = 'normal'
            this.textEle.style.fontWeight = 'bold'
            annotation.isBold = true
            annotation.isItalic = false
          } else if (fontStyle === 'italic') {
            this.textEle.style.fontStyle = 'italic'
            this.textEle.style.fontWeight = 'normal'
            annotation.isBold = false
            annotation.isItalic = true
          } else {
            annotation.isItalic = false
            annotation.isBold = false
            this.textEle.style.fontStyle = 'normal'
            this.textEle.style.fontWeight = 'normal'
          }
        }
        if (key === 'contents') {
          this.textEle.textContent = value
        }

        // Measure Properties
        if (key === 'scale' || key === 'precision') {
          if (key === 'precision') {
            this.decimalPlaces = Math.abs(annotation.precision).toString().split('.')[1]?.length || 0
          }

          await this.layer.documentViewer.annotationManager.editMeasurementAnnotation({
            rulerPtr: annotation.rulerPtr,
            [key]: value
          })
          this.getResult()
          await this.layer.documentViewer.annotationManager.editMeasurementAnnotation({
            annotPtr: annotation.annotPtr,
            contents: annotation.contents
          })
          
          const factor = await this.layer.documentViewer.messageHandler.sendWithPromise('GetFactor', annotation.rulerPtr)
          annotation.factor = factor

          if (this.textEle) {
            this.textEle.textContent = annotation.contents
          }
          if (this.layer.annotationStore.selectedElementName === this.annotation.name && !document.querySelector('.measure-panel').classList.contains('closed')) {
            this.layer.eventBus.dispatch('measurePropertyChange', this.annotation)
          }

          this.layer.eventBus.dispatch('updateMeasureData', {
            scale: value,
            length: annotation.perimeter
          })

          isUpdate = 0
        }

        if (['fontSize', 'scale', 'precision', 'contents'].includes(key)) {
          const startLine = this.svgElement.querySelector('[startLine]')
          const endLine = this.svgElement.querySelector('[endLine]')
          const dx = this.end.x - this.start.x
          const dy = this.end.y - this.start.y
          const length = Math.sqrt(dx * dx + dy * dy)
          const unitX = dx / length
          const unitY = dy / length
          const textGap = this.getTextWidth()

          if (startLine) {
            const line1End = {
              x: this.start.x + unitX * (length / 2 - textGap / 2),
              y: this.start.y + unitY * (length / 2 - textGap / 2),
            }
            startLine.setAttribute("x1", this.start.x)
            startLine.setAttribute("y1", this.start.y)
            startLine.setAttribute("x2", line1End.x)
            startLine.setAttribute("y2", line1End.y)
          }
          if (endLine) {
            const line2Start = {
              x: this.start.x + unitX * (length / 2 + textGap / 2),
              y: this.start.y + unitY * (length / 2 + textGap / 2),
            }
            endLine.setAttribute("x1", line2Start.x)
            endLine.setAttribute("y1", line2Start.y)
            endLine.setAttribute("x2", this.end.x)
            endLine.setAttribute("y2", this.end.y)
          }
        }
      }
    }
    !noSend && this.sendData({ [key]: value })
    return isUpdate
  }

  sendData(props) {
    const annotation = this.annotation
    const annotationData = {
      pagePtr: annotation.pagePtr,
      annotPtr: annotation.annotPtr,
      rulerPtr: annotation.rulerPtr,
      pageIndex: annotation.pageIndex,
      contents: annotation.contents
    }

    const isFont = ['fontSize', 'fontName', 'color', 'fontStyle'].some(item => props[item])
    if (isFont) {
      annotationData.fontSize = annotation.fontSize
      annotationData.color = annotation.color
      const fontArr = {
        Courier: {
          normal: 'Courier',
          bold: 'Courier-Bold',
          italic: 'Courier-Oblique',
          boldItalic: 'Courier-BoldOblique'
        },
        Helvetica: {
          normal: 'Helvetica',
          bold: 'Helvetica-Bold',
          italic: 'Helvetica-Oblique',
          boldItalic: 'Helvetica-BoldOblique'
        },
        Times: {
          normal: 'Times',
          bold: 'Times-Bold',
          italic: 'Times-Italic',
          boldItalic: 'Times-BoldItalic'
        }
      }

      const fontName = annotation.fontName
      if (annotation.isBold && annotation.isItalic) {
        annotationData.fontName = fontArr[fontName].boldItalic
      } else if (annotation.isBold) {
        annotationData.fontName = fontArr[fontName].bold
      } else if (annotation.isItalic) {
        annotationData.fontName = fontArr[fontName].italic
      } else {
        annotationData.fontName = fontArr[fontName] ? fontArr[fontName].normal : fontName
      }
      delete props.fontName
    }
    if (props.dashes) {
      annotationData.borderStyle = annotation.borderStyle
      annotationData.borderWidth = annotation.borderWidth
    }
    if (props.borderStyle) {
      annotationData.dashes = props.borderStyle === 'solid' ? [] : annotation.dashes
      annotationData.borderWidth = annotation.borderWidth
    }

    const updateAnnot = Object.assign({}, annotationData, props)
    this.layer.documentViewer.annotationManager.editMeasurementAnnotation(updateAnnot)
  }

  /** 计算测量结果，并更新至注释content属性上 */
  getResult() {
    // const { annotPtr, rulerPtr } = this.annotation
    // const value = await this.layer.documentViewer.messageHandler.sendWithPromise('GetMeasurementResult', { annotPtr, rulerPtr })
    // const formatValue = this.formatValue(value)

    const perimeter = 1
    this.annotation.perimeter = perimeter
    this.annotation.contents = perimeter + this.annotation.scale.translateUnit
  }

  showPopup() {
    const pageRect = this.layer.div.getBoundingClientRect()

    const left = pageRect.left + this.rect.left
    const top = this.rect.top + pageRect.top
    const right = left + this.rect.width
    const bottom = top + this.rect.height
    const rect = {
      left,
      top,
      right,
      bottom
    }
    this.layer.eventBus.dispatch('showMeasurementPopup', { rect })
    this.isPopupShow = true
  }

  hidePopup() {
    if (!this.isPopupShow) return
    this.isPopupShow = false
    this.layer.eventBus.dispatch('showMeasurementPopup')
  }

  async handlePopup(data) {
    if (data === 'delete') {
      this.handleDelete()
    }
    if (data === 'reply') {
      this.openReply()
    }
  }

  // 根据位数保留小数
  formatValue(value) {
    return value % 1 === 0 ? value : Number(value.toFixed(this.decimalPlaces))
  }

  /**
   * 获取文本宽度
   * @param {string} text - 文本
   * @param {number} lineLength - 文本所在线段的长度
   */
  getTextWidth(text, lineLength = 0) {
    const canvas = document.createElement('canvas')
    const context = canvas.getContext('2d')

    context.font = `${this.annotation.fontSize}px ${this.annotation.fontFamily}`
    const metrics = context.measureText(text)
    const gap = metrics.width * 1.1

    return lineLength > gap ? gap : lineLength
  }

  async handleKeyDown (e) {
    if (e.target.tagName === 'INPUT' || e.target.tagName === 'TEXTAREA') return
    
    let keyCode = e.keyCode || e.which
    if (keyCode === 8 || keyCode === 46) { // 8 backspace键，46 delete键
      this.handleDelete()
    }
  }

  /** 更新缩放后的rect（此rect不在注释属性上） */
  updateRect() {
    this.vertices = this.realVertices.map(item => getInitialPoint(item, this.layer.viewport.scale))
    this.annotation.rect = this.rectCalcByVertices(this.vertices)
    const { start, end } = this.getActualRect(this.layer.viewport.scale)
    this.start = start
    this.end = end
    const rect = this.calculate(start, end)
    this.rect = rect
  }

  /** 绘制注释内容 */
  draw() {
    const annotation = this.annotation
    const scale = this.layer.viewport.scale
    const actualbdwidth = (annotation.borderWidth || 2) * scale

    let dashString = ''
    const scaleValue = Math.ceil(this.ratio * scale)
    if (annotation.dashes.length === 1 || (annotation.dashes.length === 2 && annotation.dashes[0] === annotation.dashes[1])) {
      dashString = Math.ceil(annotation.dashes[0] * scaleValue) + ''
    } else if (annotation.dashes.length > 2) {
      dashString = annotation.dashes.map(dash => Math.ceil(dash * scaleValue)).join(', ')
    }
    this.svgElement.innerHTML = ''

    const [ P0, P1, P2, P3 ] = this.realVertices

    this.shapeEle = createSvg('path', {
      'stroke': annotation.borderColor,
      'stroke-width': actualbdwidth,
      'stroke-dasharray': dashString,
      'fill': 'transparent',
      'd': `M${P0.x},${P0.y} C${P1.x},${P1.y} ${P2.x},${P2.y} ${P3.x},${P3.y}`
    })

    const midpoint = {
      x: (P0.x + P3.x) / 2,
      y: (P0.y + P3.y) / 2,
    }
    const angleDeg = Math.atan2(P3.y - P0.y, P3.x - P0.x) / Math.PI * 180
    const rotateAngle = Math.abs(angleDeg) > 90 ? angleDeg - 180 : angleDeg

    this.textEle = createSvg('text', {
      fill: annotation.color || annotation.borderColor,
      'text-anchor': 'middle',
      'dominant-baseline': 'middle',
      'font-size': annotation.fontSize,
      x: midpoint.x,
      y: midpoint.y + annotation.fontSize,
      transform: `rotate(${rotateAngle} ${midpoint.x} ${midpoint.y})`
    }, {
      pointerEvents: 'none'
    })
    this.textEle.textContent = annotation.contents

    this.svgElement.append(this.shapeEle)
    this.svgElement.append(this.textEle)
  }

  /**
   * 更新绘制选中框的顶点按钮
   * @param {number} index - 指定更新某个顶点按钮的坐标，若不指定则重绘所有顶点按钮
   */
  updateCircle(index) {
    if (Number.isFinite(index) && index < this.realVertices.length - 1 && index >= 0) {
      const { x, y } = this.realVertices[index]
      const circle = this.outerLine.querySelector('circle[data-id="point' + index + '"]')
      circle.setAttribute('cx', x)
      circle.setAttribute('cy', y)
      return
    }

    this.outerLineContainer.querySelectorAll('circle').forEach(circle => circle.remove())

    for (let i = 0; i < this.realVertices.length; i++) {
      const point = this.realVertices[i]
      const circle = createSvg(
        'circle',
        {
          'data-id': 'point' + i,
          fill: '#1460F3',
          stroke: '#FFFFFF',
          'stroke-width': 2,
          cx: point.x,
          cy: point.y,
          r: 6
        }, {
          cursor: 'default'
        }
      )
      this.outerLine.append(circle)
    }
  }

  /**
   * 修改指定点的坐标
   * @param {number} index - 指定点的序号，从0开始
   * @param {number} x - 实际x坐标
   * @param {number} y - 实际y坐标
   */
  setPathPoint(index, x, y) {
    if (index > this.realVertices.length - 1 || index < 0) return

    const annotation = this.annotation

    this.realVertices[index].x = x
    this.realVertices[index].y = y
    
    if (index === 0 || index === 3) {
      this.calculateEqualPoints()
    }

    const [ P0, P1, P2, P3 ] = this.realVertices

    this.shapeEle.setAttribute('d', `M${P0.x},${P0.y} C${P1.x},${P1.y} ${P2.x},${P2.y} ${P3.x},${P3.y}`)

    const midpoint = {
      x: (P0.x + P3.x) / 2,
      y: (P0.y + P3.y) / 2,
    }
    const angleDeg = Math.atan2(P3.y - P0.y, P3.x - P0.x) / Math.PI * 180
    const rotateAngle = Math.abs(angleDeg) > 90 ? angleDeg - 180 : angleDeg

    this.textEle.setAttribute('x', midpoint.x)
    this.textEle.setAttribute('y', midpoint.y + annotation.fontSize)
    this.textEle.setAttribute('transform', `rotate(${rotateAngle} ${midpoint.x} ${midpoint.y})`)
    this.getResult()
    this.textEle.textContent = annotation.contents

    this.layer.eventBus.dispatch('updateMeasureData', {
      scale: this.annotation.scale,
      precision: this.annotation.precision,
      length: this.annotation.perimeter
    })
  }

  calculateInkPoints() {
    const [P0, P1, P2, P3] = this.vertices;
    const inkPoints = [];
    const distanceThreshold = 1;
    const increment = 0.01;
    let previousPoint = { PointX: P0.x, PointY: P0.y };
    inkPoints.push(previousPoint);
  
    for (let t = 0; t <= 1; t += increment) {
      const x = Math.pow(1 - t, 3) * P0.x + 3 * Math.pow(1 - t, 2) * t * P1.x + 3 * (1 - t) * Math.pow(t, 2) * P2.x + Math.pow(t, 3) * P3.x;
      const y = Math.pow(1 - t, 3) * P0.y + 3 * Math.pow(1 - t, 2) * t * P1.y + 3 * (1 - t) * Math.pow(t, 2) * P2.y + Math.pow(t, 3) * P3.y;
      const currentPoint = { PointX: x, PointY: y };
  
      const dx = currentPoint.PointX - previousPoint.PointX;
      const dy = currentPoint.PointY - previousPoint.PointY;
      const distance = Math.sqrt(dx * dx + dy * dy);
  
      if (distance >= distanceThreshold) {
        const numIntermediatePoints = Math.floor(distance / distanceThreshold);
        for (let i = 1; i <= numIntermediatePoints; i++) {
          const intermediatePoint = {
            PointX: previousPoint.PointX + (dx / numIntermediatePoints) * i,
            PointY: previousPoint.PointY + (dy / numIntermediatePoints) * i
          };
          inkPoints.push(intermediatePoint);
        }
        previousPoint = currentPoint;
      } else {
        previousPoint = currentPoint;
      }
    }
  
    inkPoints.push({ PointX: P3.x, PointY: P3.y }); // 确保终点属性名一致
    this.annotation.inkPointes[0] = inkPoints;
  }
  

  estimateControlPoints(inkPoints) {
    if (inkPoints.length < 4) {
      console.log("需要至少4个点来估算控制点", inkPoints)
      return
    }

    const P0 = { x: inkPoints[0].PointX, y: inkPoints[0].PointY } // 起始点
    const P3 = { x: inkPoints[inkPoints.length - 1].PointX, y: inkPoints[inkPoints.length - 1].PointY } // 终点

    // 取前面几个点和最后几个点来估算控制点
    const midIndex = Math.floor(inkPoints.length / 2)
    const P1 = {
      x: (P0.x + inkPoints[midIndex].PointX) / 2,
      y: (P0.y + inkPoints[midIndex].PointY) / 2
    }

    const P2 = {
      x: (P3.x + inkPoints[midIndex - 1].PointX) / 2,
      y: (P3.y + inkPoints[midIndex - 1].PointY) / 2
    }

    return [ P0, P1, P2, P3 ]
  }

  /** 根据起点和终点坐标，计算三等分点 */
  calculateEqualPoints() {
    const P0 = this.realVertices[0]
    const P3 = this.realVertices[3]

    const t1 = 1 / 3
    const t2 = 2 / 3

    const P1 = {
      x: P0.x + t1 * (P3.x - P0.x),
      y: P0.y + t1 * (P3.y - P0.y)
    }

    const P2 = {
      x: P0.x + t2 * (P3.x - P0.x),
      y: P0.y + t2 * (P3.y - P0.y)
    }

    this.realVertices[1] = P1
    this.realVertices[2] = P2
  }
}
