// @ts-nocheck
import { getAbsoluteCoordinate, createSvg, getInitialPoint } from '../annotation/utils'
import Circle from '../annotation/circle.js'

interface Point {
  x: number,
  y: number
}

export default class CircleCreateTool {
  documentViewer: any
  tool: any
  eventBus: any

  defaults: {
    scale: {
      base: number
      baseUnit: string
      translate: number
      translateUnit: string
    }
    precision: number
    factor: number
  }
  defaultStyles: {
    borderWidth: number
    opacity: number
    fillTransparency: number
    color: string
    borderColor: string
    borderStyle: string
    dashes: number[]
    fontName: string
    fontSize: number
    labelPosition: string
  }
  realStartPoint: Point | null = null
  realEndPoint: Point | null = null
  circle: any | null = null // 正在创建的注释类
  decimalPlaces: number
  showPerimeter: boolean
  showArea: boolean

  onMousedown = this.handleMouseDown.bind(this)
  onMousemove = this.handleMouseMove.bind(this)
  onMouseup = this.handleMouseUp.bind(this)
  handleDefaultProperty = this.setDefaults.bind(this)

  constructor(options: {
    tool: string,
    color: string,
    layer: any
  }) {
    this._tool = options.tool
    this.color = options.color
    this.layer = options.layer
    this.documentViewer = options.layer.documentViewer
    this.eventBus = options.layer.documentViewer.eventBus
    this.annotationStore = options.layer.annotationStore

    this.defaults = { // 测量属性
      scale: {
        base: 1,
        baseUnit: 'cm',
        translate: 1,
        translateUnit: 'cm',
      },
      precision: 1,
      factor: 0.0352778,
    }
    this.defaultStyles = { // 注释属性
      borderWidth: 2,
      opacity: 1,
      fillTransparency: 1,
      color: '#FF0000',
      borderColor: '#FF0000',
      borderStyle: 'solid',
      dashes: [],
      // fillColor: '#FF0000',
      fontName: 'Helvetica', // Helvetica, Helvetica-Bold, Helvetica-BoldOblique, Courier, Courier-Oblique, Times, Times-Italic
      fontSize: 14,
      fontStyle: 'regular',
      // contents: 'test',
      labelPosition: 'inside', // inside / top
    }
    this.decimalPlaces = this.defaults.precision.toString().split('.')[1]?.length || 0 // 保留小数位
    this.showPerimeter = true
    this.showArea = true

    this.init()
  }
  
  get tool () {
    return this._tool
  }

  set tool (toolType: string) {
    if (toolType === this._tool) return
    if (!toolType) {
      this.reset()
    }
    this._tool = toolType
  }

  init() {
    this.eventBus._on('defaultPropertyChanged', this.handleDefaultProperty)

    this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)

    this.layer.pageDiv.addEventListener('mousedown', this.onMousedown)
  }

  reset() {
    this.realStartPoint = null
    this.realEndPoint = null

    this.eventBus._off('defaultPropertyChanged', this.handleDefaultProperty)

    this.layer.pageDiv.removeEventListener('mousedown', this.onMousedown)
    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
  }

  async handleMouseDown(event) {
    if (this.annotationStore.selectedElementName || this.annotationStore.creating) return

    const realPoint = getAbsoluteCoordinate(this.layer.pageDiv, event) // 实际坐标
    const oriPoint = getInitialPoint(realPoint, this.layer.viewport.scale) // 原始坐标

    this.annotationStore.creating = true
    this.realStartPoint = { x: realPoint.x, y: realPoint.y }

    const left = oriPoint.x
    const top = oriPoint.y
    const right = oriPoint.x + 1
    const bottom = oriPoint.y + 1

    const rect = { left, top, right, bottom }
    const linePoints = [left, bottom, right, top]

    const annotation = {
      measure: 1,
      type: 'circle',
      measureType: 'area',
      pageIndex: this.layer.page,
      rect,
      linePoints,
      ...this.defaults,
      ...this.defaultStyles
    }
    
    this.circle = new Circle({
      annotation,
      layer: this.layer,
      highlight: this.annotationStore.highlightLink
    })

    document.addEventListener('mousemove', this.onMousemove)
    document.addEventListener('mouseup', this.onMouseup)
    // this.updateMeasureData()
  }

  async handleMouseMove(event: MouseEvent) {
    if (!this.annotationStore.creating) return

    let { x, y } = getAbsoluteCoordinate(this.layer.pageDiv, event)
    const { width, height, scale } = this.layer.viewport

    x = Math.max(0, Math.min(width, x))
    y = Math.max(0, Math.min(height, y))

    this.realEndPoint = { x, y }
    if (this.realStartPoint?.x === this.realEndPoint.x) {
      this.realEndPoint.x += 1
    }
    if (this.realStartPoint?.y === this.realEndPoint.y) {
      this.realEndPoint.y += 1
    }

    if (event.shiftKey) {
      const distanceX = Math.abs(this.realEndPoint.x - this.realStartPoint.x)
      const distanceY = Math.abs(this.realEndPoint.y - this.realStartPoint.y)
      const distance = Math.min(distanceX, distanceY)
      if (this.realStartPoint.x < this.realEndPoint.x) {
        this.realEndPoint.x = this.realStartPoint.x + distance
      } else {
        this.realEndPoint.x = this.realStartPoint.x - distance
      }
      if (this.realStartPoint.y < this.realEndPoint.y) {
        this.realEndPoint.y = this.realStartPoint.y + distance
      } else {
        this.realEndPoint.y = this.realStartPoint.y - distance
      }
    }

    this.circle.update({ start: this.circle.start, end: this.realEndPoint })
  }

  async handleMouseUp(event: MouseEvent) {
    if (this.realEndPoint && this.realStartPoint?.x !== this.realEndPoint.x && this.realStartPoint?.y !== this.realEndPoint.y) {
      const [annot] = await this.documentViewer.annotationManager.createMeasurementAnnotation(this.circle.annotation)
      this.documentViewer.annotationManager.pushAnnotationToLayer(this.circle)
    }

    this.annotationStore.creating = false
    this.realStartPoint = null
    this.realEndPoint = null
    this.circle = null

    document.removeEventListener('mousemove', this.onMousemove)
    document.removeEventListener('mouseup', this.onMouseup)
    this.documentViewer.eventBus.dispatch('annotationUpdated', { annotations: this.documentViewer.annotations })
  }

  async setDefaults(data) {
    if (data.showPerimeter !== undefined || data.showArea !== undefined) {
      this.showPerimeter = data.showPerimeter ?? this.showPerimeter
      this.showArea = data.showArea ?? this.showArea
    }

    Object.entries(data).forEach(([newKey, newValue]) => {
      Object.entries(this.defaults).forEach(([key, value]) => {
        if (newKey === key && value !== newValue && !['showPerimeter', 'showArea'].includes(newKey)) {
          this.defaults[key] = newValue

          if (key === 'precision') {
            this.decimalPlaces = newValue.toString().split('.')[1]?.length || 0
          }
        }
      })
    })

    if (this.annotationStore.selectedElementName) return

    const measureData = {
      scale: this.defaults.scale,
      precision: this.defaults.precision,
      perimeter: 0,
      area: 0
    }

    if (this.circle) {
      const { perimeter, area } = this.circle.getResult()
      measureData.perimeter = perimeter
      measureData.area = area
    }

    this.documentViewer.eventBus.dispatch('updateMeasureData', measureData)
  }
}