// @ts-nocheck
import { getAbsoluteCoordinate, createSvg, getClickPoint, getInitialPoint } from '../annotation/utils'
import Polyline from '../annotation/polyline.js'

interface Point {
  x: number,
  y: number
}

export default class PolylineCreateTool {
  documentViewer: any
  tool: any
  eventBus: any

  defaults: {
    scale: {
      base: number
      baseUnit: string
      translate: number
      translateUnit: string
    }
    precision: number
    factor: number
  }
  defaultStyles: {
    borderWidth: number
    opacity: number
    color: string
    borderColor: string
    borderStyle: string
    dashes: number[]
    head: string
    tail: string
    fontName: string
    fontSize: number
  }
  realStartPoint: Point | null = null
  realEndPoint: Point | null = null
  polyline: any | null = null // 正在创建的注释类
  decimalPlaces: number // 保留小数位

  clickTimeId: number | null = null
  isDoubleClick: boolean = false
  addBeforeMove: boolean = false

  onClick = this.handleClick.bind(this)
  onMousemove = this.handleMouseMove.bind(this)
  onDoubleClick = this.handleDoubleClick.bind(this)

  constructor(options: {
    tool: string,
    color: string,
    layer: any
  }) {
    this._tool = options.tool
    this.color = options.color
    this.layer = options.layer
    this.documentViewer = options.layer.documentViewer
    this.eventBus = options.layer.documentViewer.eventBus
    this.annotationStore = options.layer.annotationStore

    this.defaults = { // 测量属性
      scale: {
        base: 1,
        baseUnit: 'cm',
        translate: 1,
        translateUnit: 'cm',
      },
      precision: 1,
      factor: 0.0352778,
    }
    this.defaultStyles = { // 注释属性
      borderWidth: 2,
      opacity: 1,
      color: '#FF0000',
      borderColor: '#FF0000',
      borderStyle: 'solid',
      dashes: [],
      fontName: 'Helvetica', // Helvetica, Helvetica-Bold, Helvetica-BoldOblique, Courier, Courier-Oblique, Times, Times-Italic
      fontSize: 14,
      fontStyle: 'regular',
      contents: '',
    }
    this.decimalPlaces = this.defaults.precision.toString().split('.')[1]?.length || 0 // 保留小数位

    this.init()
  }
  
  get tool () {
    return this._tool
  }

  set tool (toolType: string) {
    if (toolType === this._tool) return
    if (!toolType) {
      this.reset()
    }
    this._tool = toolType
  }

  init() {
    this.eventBus._on('defaultPropertyChanged', this.handleDefaultProperty)
    
    document.removeEventListener('mousemove', this.onMousemove)
    this.layer.pageDiv.removeEventListener('click', this.onClick)
    this.layer.pageDiv.removeEventListener('dblclick', this.onDoubleClick)

    this.layer.pageDiv.addEventListener('click', this.onClick)
    this.layer.pageDiv.addEventListener('dblclick', this.onDoubleClick)
  }

  reset() {
    this.annotationStore.creating = false
    this.realStartPoint = null
    this.realEndPoint = null

    this.eventBus._off('defaultPropertyChanged', this.handleDefaultProperty)

    document.removeEventListener('mousemove', this.onMousemove)
    this.layer.pageDiv.removeEventListener('click', this.onClick)
    this.layer.pageDiv.removeEventListener('dblclick', this.onDoubleClick)
  }

  async handleClick(event: MouseEvent) {
    if (this.clickTimeId) {
      clearTimeout(this.clickTimeId)
      this.clickTimeId = null
      return
    }

    if (this.annotationStore.selectedElementName) return

    const realPoint = getAbsoluteCoordinate(this.layer.pageDiv, event) // 实际坐标
    const oriPoint = getInitialPoint(realPoint, this.layer.viewport.scale) // 原始坐标

    // 开始创建（第一个点）
    if (!this.annotationStore.creating) {
      this.annotationStore.creating = true
      this.addBeforeMove = false

      this.realStartPoint = { x: realPoint.x, y: realPoint.y }
      
      const annotation = {
        measure: 1,
        type: 'polyline',
        measureType: 'perimeter',
        pageIndex: this.layer.page,
        vertices: [{ x: oriPoint.x, y: oriPoint.y }],
        ...this.defaults,
        ...this.defaultStyles
      }

      this.polyline = new Polyline({
        annotation,
        layer: this.layer,
        highlight: this.annotationStore.highlightLink
      })

      document.addEventListener('mousemove', this.onMousemove)

      return
    }

    this.clickTimeId = setTimeout(() => {
      this.realEndPoint = { x: realPoint.x, y: realPoint.y }
      this.realStartPoint = this.realEndPoint

      this.clickTimeId = null
    }, 250)

    // 创建第二个及之后的点
    if (!realPoint || (
      Math.abs(this.realStartPoint?.x - realPoint.x) <= 5 && 
      Math.abs(this.realStartPoint?.y - realPoint.y) <= 5
    )) return

    this.polyline.setPathPoint(this.polyline.realVertices.length - 1, realPoint.x, realPoint.y)
    this.addBeforeMove = false
  }

  async handleMouseMove(event: MouseEvent) {
    if (!this.annotationStore.creating) return

    let { x, y } = getAbsoluteCoordinate(this.layer.pageDiv, event)
    const { width, height, scale } = this.layer.viewport

    x = Math.max(0, Math.min(width, x))
    y = Math.max(0, Math.min(height, y))

    this.realEndPoint = { x, y }

    if (this.realStartPoint.x === this.realEndPoint.x && this.realStartPoint.y === this.realEndPoint.y) return

    !this.addBeforeMove && this.polyline.addPathPoint(x, y)
    this.addBeforeMove = true
    this.polyline.setPathPoint(this.polyline.realVertices.length - 1, x, y)
  }

  async setDefaults(data) {
    Object.entries(data).forEach(([newKey, newValue]) => {
      Object.entries(this.defaults).forEach(([key, value]) => {
        if (newKey === key && value !== newValue) {
          this.defaults[key] = newValue

          if (key === 'precision') {
            this.decimalPlaces = newValue.toString().split('.')[1]?.length || 0
          }
        }
      })
    })

    if (this.annotationStore.selectedElementName) return

    const measureData = {
      scale: this.defaults.scale,
      precision: this.defaults.precision,
      length: 0,
      angle: 0
    }

    if (this.polyline) {
      measureData.length = this.polyline.getPerimeter()
    }

    this.documentViewer.eventBus.dispatch('updateMeasureData', measureData)
  }

  // 双击结束创建
  async handleDoubleClick(event) {
    if (this.clickTimeId) {
      clearTimeout(this.clickTimeId)
      this.clickTimeId = null
    }

    const realPoint = getAbsoluteCoordinate(this.layer.pageDiv, event) // 实际坐标

    if (realPoint && (this.realStartPoint?.x !== realPoint.x || this.realStartPoint?.y !== realPoint.y)) {
      this.polyline.setPathPoint(this.polyline.realVertices.length - 1, realPoint.x, realPoint.y)
      this.polyline.update()

      const [annot] = await this.documentViewer.annotationManager.createMeasurementAnnotation(this.polyline.annotation)
      this.documentViewer.annotationManager.pushAnnotationToLayer(this.polyline)
    }

    this.annotationStore.creating = false
    this.realStartPoint = null
    this.realEndPoint = null
    this.polyline = null

    document.removeEventListener('mousemove', this.onMousemove)
    this.documentViewer.eventBus.dispatch('annotationUpdated', { annotations: this.documentViewer.annotations })
  }
}